<!DOCTYPE html>
<html>
  <head>
    <meta charset="utf-8" />
    <meta http-equiv="X-UA-Compatible" content="chrome=1" />
    <title>
      {{ site.title | default: site.github.repository_name }} by {{
      site.github.owner_name }}
    </title>

    <link
      rel="stylesheet"
      href="{{ '/assets/css/style.css?v=' | append: site.github.build_revision | relative_url }}"
    />
    <meta name="viewport" content="width=device-width" />
    <!--[if lt IE 9]>
      <script src="//html5shiv.googlecode.com/svn/trunk/html5.js"></script>
    <![endif]-->
  </head>
  <body>
    <div class="wrapper">
      <header>
        <h1>{{ site.title | default: site.github.repository_name }}</h1>
        <p>{{ site.description | default: site.github.project_tagline }}</p>

        {% if site.github.is_project_page %}
        <p class="view">
          <a href="{{ site.github.repository_url }}"
            >View the Project on GitHub <small>{{ github_name }}</small></a
          >
        </p>
        {% endif %}

        <ul>
          <li><a href="index.html">Home</a></li>
          <li><a href="api.html">API</a></li>
          <li><a href="appenders.html">Appenders</a></li>
          <li><a href="categories.html">Categories</a></li>
          <li><a href="layouts.html">Layouts</a></li>
          <li><a href="terms.html">Terminology</a></li>
          <li><a href="faq.html">FAQ</a></li>
          <li><a href="contrib-guidelines.html">Want to help?</a></li>
          <li><a href="contributors.html">Contributors</a></li>
        </ul>

        {% if site.github.is_user_page %}
        <p class="view">
          <a href="{{ site.github.owner_url }}">View My GitHub Profile</a>
        </p>
        {% endif %} {% if site.show_downloads %}
        <ul>
          <li>
            <a href="{{ site.github.zip_url }}"
              >Download <strong>ZIP File</strong></a
            >
          </li>
          <li>
            <a href="{{ site.github.tar_url }}"
              >Download <strong>TAR Ball</strong></a
            >
          </li>
          <li>
            <a href="{{ site.github.repository_url }}"
              >View On <strong>GitHub</strong></a
            >
          </li>
        </ul>
        {% endif %}
      </header>
      <section>{{ content }}</section>
      <footer>
        {% if site.github.is_project_page %}
        <p>
          This project is maintained by
          <a href="{{ site.github.owner_url }}">{{ site.github.owner_name }}</a>
        </p>
        {% endif %}
        <p>
          <small
            >Hosted on GitHub Pages &mdash; Theme by
            <a href="https://github.com/orderedlist">orderedlist</a></small
          >
        </p>
      </footer>
    </div>
    <script src="{{ '/assets/js/scale.fix.js' | relative_url }}"></script>

    {% if site.google_analytics %}
    <script>
      (function (i, s, o, g, r, a, m) {
        i["GoogleAnalyticsObject"] = r;
        (i[r] =
          i[r] ||
          function () {
            (i[r].q = i[r].q || []).push(arguments);
          }),
          (i[r].l = 1 * new Date());
        (a = s.createElement(o)), (m = s.getElementsByTagName(o)[0]);
        a.async = 1;
        a.src = g;
        m.parentNode.insertBefore(a, m);
      })(
        window,
        document,
        "script",
        "https://www.google-analytics.com/analytics.js",
        "ga"
      );

      ga("create", "{{ site.google_analytics }}", "auto");
      ga("send", "pageview");
    </script>
    {% endif %}
  </body>
</html>
